import argparse
import os
import math
import random
import time
import queue
import numpy as np

import torch
import os
import torch.nn as nn
import transformers.modeling_utils as modeling_utils
import torch_xla.core.xla_model as xm
import torch_xla.distributed.xla_multiprocessing as xmp
import neuronx_distributed as nxd
import torch.distributed as dist


from neuronx_distributed.parallel_layers import mappings
from neuronx_distributed.parallel_layers.parallel_state import (
    get_data_parallel_rank,
    get_data_parallel_size,
    get_pipeline_model_parallel_rank,
    get_tensor_model_parallel_rank,
    initialize_model_parallel,
)
from neuronx_distributed.utils.adamw_fp32_optim_params import AdamW_FP32OptimParams
from training_utils import (
    get_param_groups_by_weight_decay,
    get_learning_rate_scheduler,
    get_sin_cos_matrix,
    create_dataset,
    create_partition,
)

# os.environ["MASTER_ADDR"] = "localhost"
# os.environ["MASTER_PORT"] = "29501"
# PyTorch/XLA DEBUG
# # Record the graph execution in the IR format
# os.environ['XLA_SAVE_TENSORS_FMT'] = 'hlo'
# os.environ['XLA_SAVE_TENSORS_FILE'] = './xla_debug/mnist.hlo'
#
# # Record the graph execution in the HLO format
# os.environ['XLA_SAVE_TENSORS_FMT'] = 'text'
# os.environ['XLA_SAVE_TENSORS_FILE'] = './xla_debug/mnist.ir'

# Show debugging VLOG for runtime and graph compilation/execution
# os.environ['TF_CPP_MIN_LOG_LEVEL'] = '0'
# os.environ['TF_CPP_VMODULE'] = 'xla_graph_executor=5,pjrt_computation_client=3'

# os.environ['NEURON_CC_FLAGS'] = os.environ.get('NEURON_CC_FLAGS', '') + ' --no_cache'
# os.environ['NEURON_CC_FLAGS'] = os.environ.get('NEURON_CC_FLAGS', '') + ' --retry_failed_compilation'


# class Layer(nn.Module):
#     def __init__(self, n_in, n_out):
#         super(Layer, self).__init__()
#         self.linear = nn.Linear(n_in, n_out)
#
#     def forward(self, x):
#         out = self.linear(x[0])
#         return (out, 1)


class SimpleLinearModel(nn.Module):
    def __init__(self, args):
        super(SimpleLinearModel, self).__init__()
        # todo: linear层数被4整除
        self.linear_1 = nn.Linear(4, 4)
        self.linear_2 = nn.Linear(4, 4)
        self.linear_3 = nn.Linear(4, 4)
        self.linear_4 = nn.Linear(4, 4)
        self.linear_5 = nn.Linear(4, 1)


    def forward(self, text):
        predict_1 = self.linear_1(text)
        predict_2 = self.linear_2(predict_1)
        predict_3 = self.linear_3(predict_2)
        predict_4 = self.linear_4(predict_3)
        predict_5 = self.linear_5(predict_4)
        model_out = predict_5.mean()
        loss = model_out * 2
        return loss, model_out


def train_simple_model(args):

    torch.manual_seed(args.seed)
    np.random.seed(args.seed)
    random.seed(args.seed)

    pipeline_cuts = create_partition(args.num_hidden_layers, args.pipeline_parallel_size)
    print(f"pipeline_cuts {pipeline_cuts}")

    nxd_config = nxd.neuronx_distributed_config(
        tensor_parallel_size=args.tensor_parallel_size,
        pipeline_parallel_size=args.pipeline_parallel_size,
        pipeline_config={
            "transformer_layer_cls": nn.Linear,
            "num_microbatches": args.num_microbatches,
            "output_loss_value_spec": (True, False), # user必填，取决于forward return的数据类型
            "input_names": ["text"],    # Dataset() key的名称
            "pipeline_cuts": pipeline_cuts,
            "trace_file_path": args.trace_file_path,
            "param_init_fn": None,
            "autowrap_modules": [mappings],
            "use_zero1_optimizer": args.use_zero1_optimizer > 0,
            "use_optimizer_wrapper": True,
        },
        optimizer_config={
            "zero_one_enabled": args.use_zero1_optimizer > 0,
            "grad_clipping": True,
            "max_grad_norm": 1.0,
        },
        sequence_parallel=False,
        activation_checkpoint_config=None,
        model_init_config=None,
    )

    def get_model(args):
        model = SimpleLinearModel(args)
        if dist.get_rank() == 0:
            print(f"model config {args}")
        return model

    dp_rank = get_data_parallel_rank()
    dp_size = get_data_parallel_size()
    tp_rank = get_tensor_model_parallel_rank()
    pp_rank = get_pipeline_model_parallel_rank()
    print(f"dp rank, dp size, tp rank, pp rank: {dp_rank}, {dp_size}, {tp_rank}, {pp_rank}")

    train_dataloader = create_dataset(
       args.vocab_size, args.training_dir, args.train_batch_size, dp_size, dp_rank, args.seed
    )
    print(f"Creating sample dataloader finished.")

    # Create NxD model
    model = nxd.initialize_parallel_model(nxd_config, get_model, args)

    print("Creating nxd model finished.")

    # weight decay: 防止模型过拟合，提高模型的泛化能力
    param_groups = get_param_groups_by_weight_decay(model)

    opt_cls = AdamW_FP32OptimParams if args.use_fp32_optimizer > 0 else torch.optim.AdamW
    optimizer = nxd.initialize_parallel_optimizer(
        nxd_config, opt_cls, param_groups, lr=args.lr, betas=(args.beta1, args.beta2), weight_decay=args.weight_decay
    )

    lr_scheduler = get_learning_rate_scheduler(optimizer, args)

    total_steps = 0
    resume_batch_idx = None
    if args.loading_step != -1:
        user_content = nxd.load_checkpoint(
            args.checkpoint_dir,
            tag=f"step_{args.loading_step}",
            model=model,
            optimizer=optimizer,
            scheduler=lr_scheduler,
        )
        if user_content is not None:
            resume_batch_idx = user_content["batch_idx"]
            total_steps = user_content["total_steps"]

    for epoch in range(0, args.max_steps):
        if torch.distributed.get_rank() == 0:
            print(f"Epoch {epoch} begins.")

        matrix_2d = torch.tensor([
            [127, 127, 127, 127],
            [127, 127, 127, 127],
            [127, 127, 127, 127],
            [127, 127, 127, 127]
        ], dtype=torch.float32)
        text = matrix_2d
        loss = model.run_train(
            text=text,
        )

        print(f"Loss: {loss}")
        optimizer.step()
        xm.mark_step()
        #
        # matrix_2d = torch.tensor([
        #     [16, 16, 16, 16, 16, 16, 16, 16],
        # ], dtype=torch.float32)
        # for idx, text in enumerate(matrix_2d):
        #     # Calling model.run_train instead of model forward to use the PP runtime
        #     loss = model.run_train(
        #         text=text
        #     )
        #     print(f"Epoch {epoch}, Data index: {idx}, Loss: {loss}")
        #     optimizer.step()
        #     # optimizer.zero_grad()
        #     # lr_scheduler.step()
        #     xm.mark_step()

    print("Training finished successfully")
    xm.mark_step()


def _mp_fn(index, args):
    train_simple_model(args)
    xm.rendezvous("_mp_fn finished") # synchronize the parallel processes


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_microbatches", type=int, default=8, help="num_microbatches")
    parser.add_argument("--tensor_parallel_size", type=int, default=8, help="tensor_parallel_size")
    parser.add_argument("--num_hidden_layers", type=int, default=-1, help="override model number of layers")
    parser.add_argument("--hidden_size", type=int, default=-1, help="override model model hidden size")
    parser.add_argument("--train_batch_size", type=int, default=16, help="batch size")
    parser.add_argument("--pipeline_parallel_size", type=int, default=1, help="PP size")
    parser.add_argument("--seq_len", type=int, default=4096, help="PP size")
    parser.add_argument("--training_dir", type=str, default=None)
    parser.add_argument("--training_config", type=str, default=None)
    parser.add_argument("--trace_file_path", type=str, default=None)
    parser.add_argument("--tb_dir", type=str, default="")
    parser.add_argument("--max_steps", type=int, default=100, help="max steps")
    parser.add_argument("--checkpoint_freq", type=int, default=100000, help="save checkpoint freq")
    parser.add_argument("--checkpoint_dir", type=str, default=None)
    parser.add_argument("--loading_step", type=int, default=-1, help="load from step, -1 means no load")
    parser.add_argument("--num_kept_checkpoint", type=int, default=-1,
                        help="number of checkpoints kept, old checkpoint will get deleted")
    parser.add_argument("--save_load_xser", type=int, default=1, help="save/load with xla serialization")


    # simple linear model
    model_config = parser.add_argument_group(title="model config", description="simple linear model")
    model_config.add_argument("--vocab_size", default=1, type=int, help="vocab size")
    model_config.add_argument("--embedding_dim", default=1, type=int, help="embedding size")


    # optimization
    opt_grp = parser.add_argument_group(title="optimization", description="arguments for optimization")
    opt_grp.add_argument("--weight_decay", default=0.01, type=float, help="weight decay")
    opt_grp.add_argument("--beta1", default=0.9, type=float, help="beta1 parameter for Adam optimizer")
    opt_grp.add_argument("--beta2", default=0.95, type=float, help="beta2 parameter for Adam optimizer")
    opt_grp.add_argument("--use_fp32_optimizer", default=0, type=int, help="use_fp32_optimizer")
    opt_grp.add_argument("--use_zero1_optimizer", default=0, type=int, help="use_zero1_optimizer")
    opt_grp.add_argument("--seed", default=1234, type=int, help="random seed")
    opt_grp.add_argument("--use_amp", default=0, type=int, help="use amp data")
    opt_grp.add_argument("--use_deferred_init", default=0, type=int, help="use torchdistx deferred initialization")
    opt_grp.add_argument("--use_meta_device_init", default=0, type=int, help="use meta device initialization")
    opt_grp.add_argument("--use_selective_checkpoint", default=0, type=int,
                         help="enable selective activation checkpointing")
    opt_grp.add_argument("--use_sequence_parallel", default=0, type=int, help="enable sequence parallelism")


    # learning rate
    lr_grp = parser.add_argument_group(title="lr", description="arguments for learning rate schedule")
    lr_grp.add_argument("--lr", type=float, default=None, help="Initial learning rate.")
    lr_grp.add_argument("--warmup_steps", type=int, default=None, help="number of warmup_steps")
    lr_grp.add_argument("--constant_steps", type=int, default=None, help="number of warmup_steps")
    lr_grp.add_argument("--min_lr", type=float, default=None,
                        help="Minumum value for learning rate. The scheduler" "clip values below this threshold.")
    lr_grp.add_argument("--logging_interval", type=int, default=1, help="number of warmup_steps")

    args, _ = parser.parse_known_args()

    if os.environ.get("XLA_USE_BF16") or os.environ.get("XLA_DOWNCAST_BF16") or args.use_amp > 0:
        modeling_utils.get_parameter_dtype = lambda x: torch.bfloat16

    if os.environ.get("WORLD_SIZE"):
        dist.init_process_group("xla")
        _mp_fn(0, args)
    else:
        xmp.spawn(_mp_fn, args=(args,))
